use log::info;
use nom::branch::alt;
use nom::bytes::streaming::is_not;
use nom::character::is_space;
use nom::character::streaming::{
    alpha1, digit1, line_ending, multispace0, multispace1, newline, none_of, not_line_ending,
};
use nom::multi::many_m_n;
use nom::sequence::{delimited, preceded, separated_pair};
use nom::{bytes::streaming::tag, IResult};
use nom::{character, number};
use sieverman::{Methods, ServerSettings};
use std::error::Error;
pub fn do_nothing_parser(input: &str) -> IResult<&str, &str> {
    Ok((input, ""))
}

fn quoted(input: &str) -> IResult<&str, &str> {
    preceded(multispace0, delimited(tag("\""), is_not("\n\""), tag("\"")))(input)
}

fn key_value(input: &str) -> IResult<&str, (&str, &str)> {
    //  note that this is really creating a function, the parser for abc
    //  vvvvv
    //         which is then called here, returning an IResult<&str, &str>
    //         vvvvv
    preceded(multispace0, separated_pair(quoted, multispace1, quoted))(input)
}

fn parse_version(input: &str) -> IResult<&str, Vec<i32>> {
    many_m_n(
        1,
        5,
        alt((
            character::complete::i32,
            preceded(character::complete::char('.'), character::complete::i32),
        )),
    )(input)
}

fn parse_methods(input: &str) -> IResult<&str, Vec<Methods>> {
    let (rem, res) = many_m_n(1, 10, alt((alpha1, preceded(multispace1, alpha1))))(input)?;
    let methods = res
        .into_iter()
        .map(|m| match m {
            "PLAIN" => Methods::Plain,
            "OAUTHBEARER" => Methods::OAuthbearer,
            any => todo!("That method is not yet implemented: {}", any),
        })
        .collect();
    Ok((rem, methods))
}

fn parse_extensions(input: &str) -> IResult<&str, Vec<String>> {
    let (rem, res) = many_m_n(1, 200, alt((is_not(" "), preceded(tag(" "), is_not(" ")))))(input)?;
    let extensions = res.into_iter().map(|m| m.to_string()).collect();
    Ok((rem, extensions))
}

fn parse_server_config(input: &str) -> IResult<&str, ServerSettings> {
    let (remain, (key, implementation)) = key_value(input)?;
    info!("Implementation:# {}", implementation);
    let (remain, (key, version)) = key_value(remain)?;
    info!("{}", version);
    let (_, version) = parse_version(version)?;
    info!("Version:# {:?}", version);
    let (remain, (key, sasl)) = key_value(remain)?;
    info!("{}", sasl);
    let (_, sasl) = parse_methods(sasl)?;
    info!("SASL:# {:?}", sasl);
    let (remain, (key, capabilities)) = key_value(remain)?;
    info!("{}", capabilities);
    let (_, capabilities) = parse_extensions(capabilities)?;
    info!("Extensions:# {:?}", &capabilities);
    let (remain, (key, notify)) = key_value(remain)?;
    info!("{}", notify);
    let (_, notify) = parse_extensions(remain)?;
    info!("Notify:# {:?}", &notify);
    let (remain, (key, max_redirects)) = key_value(remain)?;
    info!("{}", max_redirects);
    let (_, max_redirects) = character::complete::u32(max_redirects)?;
    info!("Max Redirects:# {}", max_redirects);

    IResult::Ok((
        remain,
        ServerSettings {
            implementation: implementation.to_string(),
            version,
            sasl,
            capabilities,
            notify,
            max_redirects,
        },
    ))
}

fn main() -> Result<(), Box<dyn Error>> {
    env_logger::init();
    let input = r#"
"IMPLEMENTATION" "Stalwart ManageSieve v0.4.2"
"VERSION" "1.0"
"SASL" "PLAIN OAUTHBEARER"
"SIEVE" "body comparator-elbonia comparator-i;ascii-casemap comparator-i;ascii-numeric comparator-i;octet convert copy date duplicate editheader enclose encoded-character enotify envelope envelope-deliverby envelope-dsn environment ereject extlists extracttext fcc fileinto foreverypart ihave imap4flags imapsieve include index mailbox mailboxid mboxmetadata mime redirect-deliverby redirect-dsn regex reject relational replace servermetadata spamtest spamtestplus special-use subaddress vacation vacation-seconds variables virustest"
"NOTIFY" "mailto"
"MAXREDIRECTS" "1"
"#;
    let (remaining_input, output) = parse_server_config(input)?;
    Ok(())
}
